'List commands'

from mockaioredis.util import _NOTSET

class ListCommandsMixin:
    '''List commands mixin

    Rum (some) of the Redis list commands
    '''

    async def llen(self, key):
        '''Returns the length of the list stored at key'''
        return self._redis.llen(key)


    async def lpush(self, key, value, *values):
        '''Insert specified values at the head of the list'''
        return self._redis.lpush(key, value, *values)


    async def lpop(self, key, *, encoding=_NOTSET):
        '''Pop a value off the head of the list'''
        if encoding == _NOTSET:
            encoding = self._encoding

        ret = self._redis.lpop(key)

        if encoding is None:
            return ret

        return ret.decode(encoding)


    async def rpush(self, key, value, *values):
        '''Insert specified values at the tail of the list'''
        return self._redis.rpush(key, value, *values)


    async def rpop(self, key, *, encoding=_NOTSET):
        '''Pop a value off the tail of the list'''
        if encoding == _NOTSET:
            encoding = self._encoding

        ret = self._redis.rpop(key)

        if encoding is None:
            return ret

        return ret.decode(encoding)



from mockredis import MockRedis as _MockRedis

from .generic import GenericCommandsMixin
from .hash import HashCommandsMixin
from .list import ListCommandsMixin

__all__ = ['MockRedis']

class MockRedis(GenericCommandsMixin, HashCommandsMixin, ListCommandsMixin):
    '''Fake high-level aioredis.Redis interface'''

    def __init__(self, connection=None, encoding=None):
        # Just for API compatibility
        self._conn = connection
        self._redis = _MockRedis()

        self._encoding = encoding


async def create_redis(address, *, db=None, password=None, ssl=None,
                       encoding=None, commands_factory=MockRedis,
                       loop=None):
    '''Create a fake high-level MockRedis interface

    This function is a coroutine
    '''
    return commands_factory(None, encoding=encoding)



import asyncio

from mockaioredis.util import _NOTSET

class HashCommandsMixin:

    @asyncio.coroutine
    def hset(self, key, field, value):
        return self._redis.hset(key, field, value)

    @asyncio.coroutine
    def hget(self, key, field, encoding=_NOTSET):
        if encoding == _NOTSET:
            encoding = self._encoding
        ret = self._redis.hget(key, field)

        if encoding is None:
            return ret

        if hasattr(ret, 'decode'):
            ret = ret.decode(encoding)
        return ret

    @asyncio.coroutine
    def hgetall(self, key, encoding=_NOTSET):
        if encoding == _NOTSET:
            encoding = self._encoding

        ret = self._redis.hgetall(key)
        if encoding is None:
            return ret
        new = {}
        for k, v in ret.items():
            k = k.decode(encoding)
            if hasattr(v, 'decode'):
                v = v.decode(encoding)
            new[k] = v

        return new

    @asyncio.coroutine
    def hmset(self, key, field, value, *pairs):
        if len(pairs) % 2 != 0:
            raise TypeError("length of pairs must be an even number")
        it = iter( (field, value) + pairs )
        arg_dict = dict(zip(it, it))
        return self._redis.hmset(key, arg_dict)

    @asyncio.coroutine
    def hmset_dict(self, key, *args, **kwargs):
        args_dict = args[0].copy()
        args_dict.update(kwargs)
        return self._redis.hmset(key, args_dict)

    @asyncio.coroutine
    def hmget(self, key, field, *fields, encoding=_NOTSET):
        if encoding == _NOTSET:
            encoding = self._encoding
        ret = self._redis.hmget(key, field, *fields)

        if encoding is None:
            return ret

        new = []
        for val in ret:
            if hasattr(val, 'decode'):
                val = val.decode(encoding)
            new.append(val)

        return new





'Fake aioredis.RedisPool and related functions'
import asyncio
import collections

from .util import _NOTSET
from .commands import MockRedis, create_redis


async def create_pool(address, *, db=0, password=None, ssl=None, encoding=None,
                      minsize=1, maxsize=10, commands_factory=_NOTSET, loop=None):
    if commands_factory == _NOTSET:
        commands_factory = MockRedis

    pool = MockRedisPool(address, db, password, encoding,
                         minsize=minsize, maxsize=maxsize,
                         commands_factory=commands_factory,
                         ssl=ssl, loop=loop)
    try:
        await pool._fill_free(override_min=False)
    except Exception as ex:
        pool.close()
        await pool.wait_closed()
        raise

    return pool


class MockRedisPool:
    '''Imitate a aioredis.RedisPool

    Or at least enough of it to create, use and close a pool
    '''
    def __init__(self, address, db=0, password=0, encoding=None,
                 *, minsize, maxsize, commands_factory, ssl=None, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()

        self._address = address
        self._db = db
        self._password = password
        self._encoding = encoding
        self._minsize = minsize
        self._maxsize = maxsize
        self._factory = commands_factory
        self._ssl = ssl
        self._loop = loop

        # fake it here, we always only have one connection
        self._pool = collections.deque(maxlen=1)
        self._used = set()
        self._acquiring = 0

        self._cond = asyncio.Condition(loop=loop)
        self._close_state = asyncio.Event(loop=loop)
        self._close_waiter = asyncio.ensure_future(self._do_close(), loop=loop)


    @property
    def minsize(self):
        '''always return 1'''
        return 1


    @property
    def maxsize(self):
        '''always return 1'''
        return 1


    @property
    def size(self):
        return self.freesize + len(self._used) + self._acquiring


    @property
    def freesize(self):
        return len(self._pool)


    async def _do_close(self):
        await self._close_state.wait()
        async with self._cond:
            waiters = []
            while self._pool:
                conn = self._pool.popleft()
            # fake connections, so no need to do anything for used connections


    def close(self):
        if not self._close_state.is_set():
            self._close_state.set()


    @property
    def closed(self):
        return self._close_state.is_set()


    async def wait_closed(self):
        'wait until pool is closed'
        await asyncio.shield(self._close_waiter, loop=self._loop)


    async def acquire(self):
        '''Pretend to aquire a connection.

        In fact, always return the same MockRedis object once free'''
        async with self._cond:
            while True:
                await self._fill_free(override_min=True)
                if self.freesize:
                    conn = self._pool.popleft()
                    self._used.add(conn)
                    return conn
                else:
                    await self._cond.wait()


    def release(self, conn):
        '''Release our single MockRedis connection'''
        assert conn in self._used, "Invalid connection, maybe from other pool?"
        self._used.remove(conn)
        self._pool.append(conn)
        asyncio.ensure_future(self._wakeup(), loop=self._loop)


    async def _fill_free(self, *, override_min):
        while self.size < self.minsize:
            self._acquiring += 1
            try:
                conn = await self._create_new_connection()
                self._pool.append(conn)
            finally:
                self._acquiring -= 1
        if self.freesize:
            return
        if override_min:
            while not self._pool and self.size < self.maxsize:
                self._acquiring += 1
                try:
                    conn = await self._create_new_connection()
                    self._pool.append(conn)
                finally:
                    self._acquiring -= 1


    def _create_new_connection(self):
        return create_redis(self._address,
                            db=self._db,
                            password=self._password,
                            ssl=self._ssl,
                            encoding=self._encoding,
                            commands_factory=self._factory,
                            loop=self._loop)


    async def _wakeup(self):
        async with self._cond:
            self._cond.notify()


    def get(self):
        '''Return async context manager for working with the connection

        async with pool.get() as conn:
            await conn.get(key)
        '''
        return _AsyncConnectionContextManager(self)


class _AsyncConnectionContextManager:

    __slots__ = ('_pool', '_conn')

    def __init__(self, pool):
        self._pool = pool
        self._conn = None

    async def __aenter__(self):
        self._conn = await self._pool.acquire()
        return self._conn

    async def __aexit__(self, exc_type, exc_value, tb):
        try:
            self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None





'Generic Redis commands'


class GenericCommandsMixin:
    '''Generic commands mixin
    '''

    async def delete(self, key, *keys):
        '''Delete specified key(s)'''
        return self._redis.delete(key, *keys)


    async def exists(self, key, *keys):
        '''Check if key(s) exist

        Like in Redis 3.0.3+, returns int count of existing keys.
        If the same existing key is given multiple times, it is
        counted multiple times.
        '''
        all_keys = (key, ) + keys
        existing = 0
        # redis-py and, by extension mockredispy still only support
        # single-key checks that return True/False
        for k in all_keys:
            if self._redis.exists(k):
                existing += 1

        return existing



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

install_requires = [
    'mockredispy-kblin'
]

tests_require = [
    'coverage',
    'pytest',
    'pytest-asyncio',
    'pytest-cov',
]

classifiers = [
    'License :: OSI Approved :: Apache Software License',
    'Development Status :: 3 - Alpha',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Operating System :: POSIX',
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
]


def read_version():
    import re
    import os.path
    regexp = re.compile(r"^__version__\W*=\W*'([\d.abrc]+)'")
    init_py = os.path.join(os.path.dirname(__file__),
                           'mockaioredis', '__init__.py')
    with open(init_py, 'r') as fh:
        for line in fh:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        else:
            raise RuntimeError('Unable to find version in ' + init_py)


def read(filename):
    with open(filename, 'r') as fh:
        return fh.read()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import sys
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


setup(name='mockaioredis',
      version=read_version(),
      description="Mock implementation of aioredis",
      long_description=read('README.rst'),
      classifiers=classifiers,
      platforms=["POSIX"],
      author="Kai Blin",
      author_email="kblin@biosustain.dtu.dk",
      url="https://github.com/kblin/mockaioredis",
      license="Apache Software License",
      packages=find_packages(exclude=["tests"]),
      install_requires=install_requires,
      tests_require=tests_require,
      cmdclass={'test': PyTest},
      include_package_data=True,
      extras_require={
        'testing': tests_require,
      },
)
